/*
 *   tunpipe       TUN device <--> data stream converter
 *   Copyright (c) 2006 Michal Trojnara <Michal.Trojnara@mirt.net>
 *                 All Rights Reserved
 *
 *   Version:      1.00             (tunpipe.c)
 *   Date:         2006.02.02
 *
 *   Author:       Michal Trojnara  <Michal.Trojnara@mirt.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <syslog.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <linux/if_tun.h>
#include <pthread.h>
#include <signal.h>

/**************************************** data structures */

#define MTU 8192
#define LEN sizeof(u_short)

typedef struct {
    u_short len;
    char buf[MTU];
} PKT;

typedef union {
    PKT pkt;
    char buf[sizeof(PKT)];
} BUF;

/**************************************** global variables */

static int log, signal_pipe[2], tunfd;
volatile static int pkt_in=0, pkt_out=0, bytes_in=0, bytes_out=0;
static pthread_t log_thread;

/**************************************** logging */

static void *log_proc(void *param) {
    char buf[256], *nl;
    int len, num;

    len=0;
    while((num=read(log, buf+len, sizeof(buf)-len-1))>0) {
        buf[len+num]='\0'; /* build a null-terminated string */
        len=strlen(buf); /* in case '\0' is logged */
        while((nl=index(buf, '\n'))!=NULL) { /* a new line */
            *nl='\0';
            syslog(LOG_ERR, "%s", buf);
            len-=nl-buf+1;
            memmove(buf, nl+1, len);
        }
    }
    return NULL;
}

static void log_exit(void) {
    fflush(stderr); /* just in case buffers are not empty */
    close(2); /* send EOF to the pipe */
    pthread_join(log_thread, (void **)NULL);
}

static void log_init(void) {
    int log_pipe[2];
    sigset_t newmask, oldmask;

    pipe(log_pipe);
    log=log_pipe[0]; /* setup global variable */
    openlog("tunpipe", LOG_CONS, LOG_DAEMON);
    sigfillset(&newmask);
    pthread_sigmask(SIG_SETMASK, &newmask, &oldmask); /* block signals */
    pthread_create(&log_thread, NULL, log_proc, NULL);
    pthread_sigmask(SIG_SETMASK, &oldmask, NULL); /* restore the mask */
    atexit(log_exit);
    dup2(log_pipe[1], 2); /* dup 2 stderr */
    close(log_pipe[1]); /* no longer needed */
}

/**************************************** signals */

static void signal_handler(int sig) {
    write(signal_pipe[1], &sig, sizeof(int));
}

static void signal_init() {
    pipe(signal_pipe);
    signal(SIGPIPE, SIG_IGN); /* avoid 'broken pipe' signal */
    signal(SIGUSR1, signal_handler);
    signal(SIGTERM, signal_handler);
    signal(SIGQUIT, signal_handler);
    signal(SIGINT, signal_handler);
    signal(SIGHUP, signal_handler);
}

static void signal_process() {
    int sig;

    while(1) {
        while(read(signal_pipe[0], &sig, sizeof(int))!=sizeof(int))
            ; /* do nothing */
        switch(sig) {
        case SIGUSR1:
            fprintf(stderr,
                "in: %d packet(s) %d byte(s), out: %d packet(s), %d byte(s)\n",
                pkt_in, bytes_in, pkt_out, bytes_out);
            break;
        default:
            fprintf(stderr, "Received signal %d, terminating\n", sig);
            exit(0);
        }
    }
}

/**************************************** TUN */

static void *tun_read(void *param) {
    BUF buf;
    int num;

    while(1) {
        num=read(tunfd, &buf.pkt.buf, MTU);
        switch(num) {
        case -1:
            perror("Reading TUN");
            exit(1);
        case 0:
            fprintf(stderr, "TUN closed\n");
            exit(1);
        }
        bytes_in+=num;
        buf.pkt.len=htons(num);
        num=write(1, &buf.buf, LEN+num);
        if(num<0) {
            perror("Writing PIPE");
            exit(1);
        }
        ++pkt_in;
    }
    return NULL;
}

static void *tun_write(void *param) {
    BUF buf;
    int len, num;

    len=0;
    while(1) {
        num=read(0, &buf.buf, sizeof(buf));
        switch(num) {
        case -1:
            perror("Reading PIPE");
            exit(1);
        case 0:
            fprintf(stderr, "PIPE closed\n");
            exit(1);
        }
        len+=num;
        while(len>=LEN && len>=LEN+ntohs(buf.pkt.len)) {
            num=write(tunfd, &buf.pkt.buf, ntohs(buf.pkt.len));
            if(num<0) {
                perror("Writing TUN");
                exit(1);
            }
            bytes_out+=num;
            num=LEN+ntohs(buf.pkt.len);
            len-=num;
            memmove(&buf.buf, &buf.buf+num, len);
            ++pkt_out;
        }

    }
    return NULL;
}

static void tun_init(char *script) {
    struct ifreq ifr;
    char cmd[256];
    int tmpfd, err;
    pthread_t tun_thread_read, tun_thread_write;
    sigset_t newmask, oldmask;

    tunfd=open("/dev/net/tun", O_RDWR);
    if(tunfd<0) {
        perror("open: /dev/net/tun");
        exit(1);
    }
    fcntl(tunfd, F_SETFD, FD_CLOEXEC);
    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = IFF_TUN;
    if(ioctl(tunfd, TUNSETIFF, (void *)&ifr)<0) {
        perror("ioctl: /dev/net/tun");
        exit(1);
    }
    snprintf(cmd, sizeof(cmd), "%s %s\n", script, ifr.ifr_name);
    tmpfd=dup(1); /* save copy of stdout descriptor */
    fcntl(tmpfd, F_SETFD, FD_CLOEXEC);
    dup2(2, 1); /* redirect stdout to stderr */
    err=system(cmd);
    dup2(tmpfd, 1); /* restore stdout */
    close(tmpfd); /* close temporary descriptor */
    if(err) {
        fprintf(stderr, "system() failed\n");
        exit(1);
    }
    sigfillset(&newmask);
    pthread_sigmask(SIG_SETMASK, &newmask, &oldmask); /* block signals */
    pthread_create(&tun_thread_read, NULL, tun_read, NULL);
    pthread_create(&tun_thread_write, NULL, tun_write, NULL);
    pthread_sigmask(SIG_SETMASK, &oldmask, NULL); /* restore the mask */
}

/**************************************** main() function */

int main(int argc, char *argv[]) {
    signal_init();
    log_init();
    if(argc<2) {
        fprintf(stderr, "Interface configuration script not specified\n");
        exit(1);
    }
    tun_init(argv[1]);
    fprintf(stderr, "tunpipe started\n");
    signal_process();
    return 0;
}

/**************************************** end of tunpipe.c file */
